import DeepProxy from 'proxy-deep';
import { useQueryClient, createQuery, createMutation, createInfiniteQuery, createQueries, } from '@tanstack/svelte-query';
import { onDestroy } from 'svelte';
function getArrayQueryKey(queryKey, input, type) {
    const arrayPath = (typeof queryKey === 'string' ?
        queryKey === '' ? [] : queryKey.split('.')
        : queryKey);
    if (!input && (!type || type === 'any'))
        // for `utils.invalidate()` to match all queries (including vanilla react-query)
        // we don't want nested array if path is empty, i.e. `[]` instead of `[[]]`
        return arrayPath.length ? [arrayPath] : [];
    return [
        arrayPath,
        {
            ...(typeof input !== 'undefined' && { input: input }),
            ...(type && type !== 'any' && { type: type }),
        },
    ];
}
// getContext
const ContextProcedureNames = {
    client: 'client',
    fetch: 'fetch',
    prefetch: 'prefetch',
    fetchInfinite: 'fetchInfinite',
    prefetchInfinite: 'prefetchInfinite',
    invalidate: 'invalidate',
    refetch: 'refetch',
    reset: 'reset',
    cancel: 'cancel',
    setData: 'setData',
    getData: 'getData',
    setInfiniteData: 'setInfiniteData',
    getInfiniteData: 'getInfiniteData',
};
// Procedures
const ProcedureNames = {
    query: 'useQuery',
    infiniteQuery: 'useInfiniteQuery',
    mutate: 'useMutation',
    subscribe: 'useSubscription',
    queryKey: 'getQueryKey',
    context: 'useContext',
    queries: 'useQueries',
};
// Implementation
function useSubscription(cb, input, opts) {
    const enabled = opts?.enabled ?? true;
    if (!enabled)
        return;
    let isStopped = false;
    const subscription = cb(input, {
        onStarted: () => {
            if (!isStopped)
                opts.onStarted?.();
        },
        onData: (data) => {
            if (!isStopped)
                opts.onData?.(data);
        },
        onError: (err) => {
            if (!isStopped)
                opts.onError?.(err);
        }
    });
    return onDestroy(() => {
        isStopped = true;
        subscription.unsubscribe();
    });
}
function createUseQueriesProxy(client) {
    return new DeepProxy({}, {
        get() {
            return this.nest(() => { });
        },
        apply(_target, _thisArg, argList) {
            const target = [...this.path].reduce((client, value) => client[value], client);
            const [input, opts] = argList;
            return {
                ...opts,
                queryKey: getArrayQueryKey(this.path, input, 'query'),
                queryFn: () => target.query(input),
            };
        }
    });
}
function createUseContextProxy(client, queryClient) {
    return new DeepProxy({}, {
        get(_target, key, _receiver) {
            if (key === ContextProcedureNames.client)
                return client;
            return this.nest(() => { });
        },
        apply(_target, _thisArg, argList) {
            const utilName = this.path.pop();
            // TODO: should probably replace `reduce` with `for...of` for better performance
            const target = [...this.path].reduce((client, value) => client[value], client);
            const [input, ...rest] = argList;
            if (utilName === ContextProcedureNames.invalidate) {
                const [filters] = rest;
                return queryClient.invalidateQueries({
                    ...filters,
                    queryKey: getArrayQueryKey(this.path, input, 'any')
                });
            }
            else if (utilName === ContextProcedureNames.fetch) {
                const [opts] = rest;
                return queryClient.fetchQuery({
                    ...opts,
                    queryKey: getArrayQueryKey(this.path, input, 'query'),
                    queryFn: () => target.query(input),
                });
            }
            else if (utilName === ContextProcedureNames.fetchInfinite) {
                const [opts] = rest;
                return queryClient.fetchInfiniteQuery({
                    ...opts,
                    queryKey: getArrayQueryKey(this.path, input, 'infinite'),
                    queryFn: ({ pageParam }) => target.query({ ...input, cursor: pageParam }),
                });
            }
            else if (utilName === ContextProcedureNames.prefetch) {
                const [opts] = rest;
                return queryClient.prefetchQuery({
                    ...opts,
                    queryKey: getArrayQueryKey(this.path, input, 'query'),
                    queryFn: () => target.query(input),
                });
            }
            else if (utilName === ContextProcedureNames.prefetchInfinite) {
                const [opts] = rest;
                return queryClient.prefetchInfiniteQuery({
                    ...opts,
                    queryKey: getArrayQueryKey(this.path, input, 'infinite'),
                    queryFn: ({ pageParam }) => target.query({ ...input, cursor: pageParam }),
                });
            }
            else if (utilName === ContextProcedureNames.refetch) {
                return queryClient.refetchQueries(getArrayQueryKey(this.path, input, 'any'), ...rest);
            }
            else if (utilName === ContextProcedureNames.reset) {
                return queryClient.resetQueries(getArrayQueryKey(this.path, input, 'any'), ...rest);
            }
            else if (utilName === ContextProcedureNames.cancel) {
                return queryClient.cancelQueries(getArrayQueryKey(this.path, input, 'any'), ...rest);
            }
            else if (utilName === ContextProcedureNames.setData) {
                return queryClient.setQueryData(getArrayQueryKey(this.path, input, 'query'), ...rest);
            }
            else if (utilName === ContextProcedureNames.getData) {
                return queryClient.getQueryData(getArrayQueryKey(this.path, input, 'query'), ...rest);
            }
            else if (utilName === ContextProcedureNames.setInfiniteData) {
                return queryClient.setQueryData(getArrayQueryKey(this.path, input, 'infinite'), ...rest);
            }
            else if (utilName === ContextProcedureNames.getInfiniteData) {
                return queryClient.getQueryData(getArrayQueryKey(this.path, input, 'infinite'), ...rest);
            }
            // Just simulating the error thrown by `@trpc/react-query` for now.
            throw new TypeError('contextMap[utilName] is not a function');
        }
    });
}
export function svelteQueryWrapper(client, queryClient) {
    const useQueriesProxy = createUseQueriesProxy(client);
    return new DeepProxy({}, {
        get() {
            return this.nest(() => { });
        },
        apply(_target, _thisArg, argList) {
            const procedure = this.path.pop();
            // TODO: should probably replace `reduce` with `for...of` for better performance
            const target = [...this.path].reduce((client, value) => client[value], client);
            const [input, opts] = argList;
            // NOTE: should probably throw error for procedures
            // for procedures with conflicting names like `useQuery`,
            // `useMutation`, etc but it is not handled in `@trpc/react-query` either.
            if (procedure === ProcedureNames.query) {
                return createQuery({
                    ...opts,
                    queryKey: getArrayQueryKey(this.path, input, 'query'),
                    queryFn: () => target.query(input),
                });
            }
            else if (procedure === ProcedureNames.infiniteQuery) {
                return createInfiniteQuery({
                    ...opts,
                    queryKey: getArrayQueryKey(this.path, input, 'infinite'),
                    queryFn: ({ pageParam }) => target.query({ ...input, cursor: pageParam }),
                });
            }
            else if (procedure === ProcedureNames.mutate) {
                return createMutation({
                    ...opts,
                    mutationKey: this.path,
                    mutationFn: (data) => target.mutate(data),
                });
            }
            else if (procedure === ProcedureNames.subscribe) {
                return useSubscription(target.subscribe, input, opts);
            }
            else if (procedure === ProcedureNames.queryKey) {
                // NOTE: should probably throw error for procedures
                // like `useMutation` that don't have `getQueryKey`,
                // but it is not handled in `@trpc/react-query` either.
                return getArrayQueryKey(this.path, input, opts);
            }
            else if (procedure === ProcedureNames.queries) {
                if (this.path.length === 0) {
                    return createQueries(input(useQueriesProxy));
                }
            }
            else if (procedure === ProcedureNames.context) {
                return createUseContextProxy(client, queryClient ? queryClient : useQueryClient());
            }
            return target[procedure].call();
        }
    });
}
;
